@@ -2,7 +2,9 @@ from typing import Tuple, Dict, Any
 from web3 import HTTPProvider, Web3
 from eth_utils.abi import function_abi_to_4byte_selector, collapse_if_tuple
 from actions_lib.actions.type import Action, ActionData
+from actions_lib.actions.user.user import decrease_token_amount, increase_token_amount
 from actions_lib.utils.contact_tool import get_contact_address
+from cdp import *
 
 # Constants
 TOKEN_MAPPING = {
@@ -11,10 +13,16 @@ TOKEN_MAPPING = {
             "address": None,
             "decimal": 18
         },
+        # testnet
         "usdc": {
             "address": "0xd10519Aa03FE7Ffb95189B52B74F94Fb33E2C88a",
             "decimal": 6
         },
+        # mainnet
+        # "usdc": {
+        #     "address": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
+        #     "decimal": 6
+        # },
         "usdt": {
             "address": "0xdAC17F958D2ee523a2206206994597C13D831ec7",
             "decimal": 6
@@ -35,6 +43,14 @@ TOKEN_MAPPING = {
             "address": "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
             "decimal": 8
         },
+        "wbtc": {
+            "address": "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
+            "decimal": 8
+        },
+        "cbbtc": {
+            "address": "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
+            "decimal": 8
+        },
         "ens": {
             "address": "0xC18360217D8F7Ab5e7c516566761Ea12Ce7F9D72",
             "decimal": 18
@@ -45,10 +61,16 @@ TOKEN_MAPPING = {
             "address": None,
             "decimal": 18
         },
+        # testnet
         "usdc": {
             "address": "0xd10519Aa03FE7Ffb95189B52B74F94Fb33E2C88a",
             "decimal": 6
         },
+        # mainnet
+        # "usdc": {
+        #     "address": "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
+        #     "decimal": 6
+        # },
         "usdt": {
             "address": "0xdAC17F958D2ee523a2206206994597C13D831ec7",
             "decimal": 6
@@ -87,10 +109,16 @@ TOKEN_MAPPING = {
             "address": None,
             "decimal": 18
         },
+        # testnet
         "usdc": {
             "address": "0x45b58118a5374dccf7fd6f3f66c66278ab7473d9",
             "decimal": 6
         },
+        # mainnet
+        # "usdc": {
+        #     "address": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
+        #     "decimal": 6
+        # },
         "btc": {
             "address": "0xcbB7C0000aB88B473b1f5aFd9ef808440eed33Bf",
             "decimal": 8
@@ -112,6 +140,12 @@ CHAIN_ID_MAPPING = {
     'eth': '0xaa36a7'
 }
 
+# CHAIN_ID_MAPPING = {
+#     'base': '0x2105',
+#     'ethereum': '0x1',
+#     'eth': '0x1'
+# }
+
 # ABIs
 ALLOWANCE_ABI = """[{"constant":true,"inputs":[{"name":"_owner","type":"address"},{"name":"_spender","type":"address"}],"name":"allowance","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"}]"""
 
@@ -140,6 +174,57 @@ ERC20_APPROVE_ABI = """[{"constant":false,"inputs":[{"name":"_spender","type":"a
 
 ERC20_TRANSFER_ABI = """[{"constant":false,"inputs":[{"name":"recipient","type":"address"},{"name":"amount","type":"uint256"}],"name":"transfer","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"}]"""
 
+def mpc_transfer(default_address, destination, chain_raw, token_raw, token_address, amount):
+    try:
+        if chain_raw == 'base' and token_raw == 'usdc':
+            transfer = default_address.transfer(amount, 'usdc', destination, gasless=True)
+            transfer.wait()
+            transferred_wei = convert_to_wei(float(amount), chain_raw, token_raw)
+            if transfer.status is Transaction.Status.COMPLETE:
+                status = 200
+                transfer_status = 1
+                print(f"Transfer complete: {transfer.transaction_hash}")
+                print(f"Transfer successfully landed on-chain: {transfer.transaction_link}")
+            else:
+                amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
+                # increase_token_amount(redis_client, executor, chain_raw, token_raw, amount_to_transfer)
+                status = 500
+                transfer_status = 0
+                transferred_wei = 0
+                print(f"Transfer failed on-chain: {transfer.transaction_link}")
+
+            info = {
+                "content": generate_tx_link_content(transfer.status, transfer.transaction_link),
+                "tx_hash": transfer.transaction_hash,
+                "chain": chain_raw,
+                "status": transfer_status,
+                "code": status,
+                "transferred_wei": transferred_wei
+            }
+
+            return info
+        else:
+            # amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
+            # transfer = default_address.invoke_contract(
+            #     contract_address=token_address,
+            #     abi=ERC20_TRANSFER_ABI,
+            #     method="transfer",
+            #     args={"to": destination, "value": amount_to_transfer}
+            # )
+            # increase_token_amount(redis_client, executor, chain_raw, token_raw, amount_to_transfer)
+            info = {
+                "content": f"In automatic mode, only USDC transfers on the Base mainnet are supported.",
+                "code": 500
+            }
+            return info
+    except Exception as e:
+        info = {
+            "content": f"{e}",
+            "code": 500
+        }
+        # increase_token_amount(redis_client, executor, chain_raw, token_raw, amount_to_transfer)
+        return info
+
 def get_web3(providers, chain):
     if providers[chain]['w3'] is None:
         providers[chain]['w3'] = Web3(HTTPProvider(providers[chain]['rpc']))
@@ -147,7 +232,7 @@ def get_web3(providers, chain):
 
 def get_user_action_mode(redis_client, user_id):
     user_id = user_id.lower()
-    action_mode_key = f"user:{user_id}:action_mode"
+    action_mode_key = f"user:action_mode:{user_id}"
     mode = redis_client.get(action_mode_key)
     if mode:
         return mode
@@ -187,6 +272,8 @@ def transfer(receiver: str, amount: float, step: any, chain: str = "base", token
     eth_w3 = get_web3(providers, 'eth')
     token_raw = token.lower()
     token_address = TOKEN_MAPPING[chain_raw][token_raw]['address']
+
+    mpc_address = kwargs.get('mpc_address')
     if token_address is not None:
         token_address = w3.to_checksum_address(token_address)
     print(f"token address: {token_address}", flush=True)
@@ -194,11 +281,11 @@ def transfer(receiver: str, amount: float, step: any, chain: str = "base", token
     print(f"w3: {w3}", flush=True)
 
     if run_mode.lower() == 'auto':
-        return handle_auto_transfer(w3, eth_w3, executor, receiver, amount, token_address, spender_address, sender_address, control_contract, private_key, redis_client, chain_raw, token_raw, blockchain_explore, executor_balance)
+        return handle_auto_transfer(w3, eth_w3, mpc_address, executor, receiver, amount, token_address, spender_address, sender_address, control_contract, private_key, redis_client, chain_raw, token_raw, blockchain_explore, executor_balance)
     else:
         return handle_manual_transfer(w3, eth_w3, executor, receiver, amount, token_address, redis_client, chain_raw, token_raw)
 
-def handle_auto_transfer(w3: Web3, eth_w3, executor, receiver, amount, token_address, spender_address, sender_address, control_contract, private_key, redis_client, chain_raw, token_raw, blockchain_explore, executor_balance):
+def handle_auto_transfer(w3: Web3, eth_w3, mpc_address, executor, receiver, amount, token_address, spender_address, sender_address, control_contract, private_key, redis_client, chain_raw, token_raw, blockchain_explore, executor_balance):
     if token_raw != 'usdc':
         token = token_raw.upper()
         if token_raw in ['btc', 'cbbtc', 'wbtc']:
@@ -208,44 +295,56 @@ def handle_auto_transfer(w3: Web3, eth_w3, executor, receiver, amount, token_add
                 token = 'WBTC'
         return {'result': {"code": 500, "content": f"Transfers of the current currency({token}) are not supported in automatic mode."}, 'action': None, 'next_action': None}
     else:
-        token_contract = w3.eth.contract(address=token_address, abi=ALLOWANCE_ABI)
-        allowance = token_contract.functions.allowance(Web3.to_checksum_address(executor), spender_address).call()
-        allowance_normalized = convert_to_value(allowance, chain_raw, token_raw)
-        print(f"allowance_normalized: {float(allowance_normalized)}")
+        # token_contract = w3.eth.contract(address=token_address, abi=ALLOWANCE_ABI)
+        # allowance = token_contract.functions.allowance(Web3.to_checksum_address(executor), spender_address).call()
+        # allowance_normalized = convert_to_value(allowance, chain_raw, token_raw)
+        # print(f"allowance_normalized: {float(allowance_normalized)}")
+        amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
+        try:
+            decrease_token_amount(redis_client, executor, chain_raw, token, amount_to_transfer)
+        except Exception as e:
+            print(f"Exception occurred: {e}", flush=True)
+            return {'result': {"content": f"{e}", "code": 500}, 'action': None, 'next_action': None}
         
-        if float(allowance_normalized) < float(amount):
-            return handle_insufficient_allowance(w3, eth_w3, executor, receiver, amount, spender_address, token_address, chain_raw, token_raw, redis_client)
+        # if float(allowance_normalized) < float(amount):
+        #     return handle_insufficient_allowance(w3, eth_w3, executor, receiver, amount, spender_address, token_address, chain_raw, token_raw, redis_client)
 
         code, receiver_address = get_contact_address(redis_client, executor, receiver, eth_w3)
         if code != 200:
-            return {'result': {"content": f"Cannot find {receiver} on contact", "code": 420}, 'action': None, 'next_action': None}
+            return {'result': {"content": f"Cannot find {receiver} on contacts", "code": 420}, 'action': None, 'next_action': None}
 
-        amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
+        # amount_to_transfer = convert_to_wei(float(amount), chain_raw, token_raw)
         to = w3.to_checksum_address(receiver_address)
         executor = w3.to_checksum_address(executor)
 
         print(f"to: {to}, executor: {executor}")
         
         try:
-            can_send, tx_hash = maybe_send_transaction(w3, control_contract, sender_address, token_address, executor, to, amount_to_transfer, private_key, chain_raw, executor_balance)
-            if can_send:
-                tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
-                status = 200 if tx_receipt['status'] == 1 else 500
-                info = {
-                    "content": generate_markdown_content(tx_receipt['status'], tx_hash.hex(), blockchain_explore),
-                    "tx_hash": tx_hash.hex(),
-                    "chain": chain_raw,
-                    "status": tx_receipt['status'],
-                    "code": status
-                }
-                return {'result': info, 'action': None, 'next_action': None}
-            else:
-                return {'result': {"code": 402, "content": "Your account balance is insufficient, please top up first or apply to join the credit list for getting free $5 funding."}, 'action': None, 'next_action': None}
+            info = mpc_transfer(mpc_address, to, chain_raw, token_raw, token_address, amount)
+
+            if info['code'] != 200:
+                increase_token_amount(redis_client, executor, chain_raw, token, amount_to_transfer)
+
+            return {'result': info, 'action': None, 'next_action': None}
+            # can_send, tx_hash = maybe_send_transaction(w3, control_contract, sender_address, token_address, executor, to, amount_to_transfer, private_key, chain_raw, executor_balance)
+            # if can_send:
+            #     tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
+            #     status = 200 if tx_receipt['status'] == 1 else 500
+            #     info = {
+            #         "content": generate_markdown_content(tx_receipt['status'], tx_hash.hex(), blockchain_explore),
+            #         "tx_hash": tx_hash.hex(),
+            #         "chain": chain_raw,
+            #         "status": tx_receipt['status'],
+            #         "code": status
+            #     }
+            #     return {'result': info, 'action': None, 'next_action': None}
+            # else:
+            #     return {'result': {"code": 402, "content": "Your account balance is insufficient. Please top up first, or apply to join the credit list to receive $5 of Ardio Credit."}, 'action': None, 'next_action': None}
         except Exception as e:
             return handle_transaction_error(e, control_contract, w3)
 
 def handle_manual_transfer(w3: Web3, eth_w3, executor, receiver, amount, token_address, redis_client, chain_raw, token_raw):
-    active_chain_key = f"user:{executor}:active_chain"
+    active_chain_key = f"user:active_chain:{executor}"
     active_chain = redis_client.get(active_chain_key) or "base"
     print(f"active_chain_key: {active_chain_key}, active_chain: {active_chain}", flush=True)
     
@@ -325,10 +424,11 @@ def handle_incorrect_network(w3, eth_w3, executor, receiver, amount, chain_raw,
     action_data = ActionData(func='switch_chain', group='', params=params)
     action = Action(msg=None, type='setting', data=action_data.__dict__)
 
-    code, receiver_address = get_contact_address(redis_client, executor, receiver, eth_w3)
+    code, get_contact_res = get_contact_address(redis_client, executor, receiver, eth_w3)
     if code != 200:
-        return {'result': {"content": f"Cannot find {receiver} on contact", "code": 420}, 'action': None, 'next_action': None}
+        return {'result': {"content": get_contact_res, "code": code}, 'action': None, 'next_action': None}
 
+    receiver_address = get_contact_res
     next_params = {
         'receiver': w3.to_checksum_address(receiver_address),
         'amount': amount,
@@ -375,6 +475,12 @@ def generate_markdown_content(status, hash_value, blockchain_explore):
     status_text = "successful" if status == 1 else "failed"
     return f"Transfer {status_text}. You can check the transaction on [blockchain explorer]({blockchain_explore}{hash_value})"
 
+
+def generate_tx_link_content(status, tx_link):
+    status_text = "successful" if status is Transaction.Status.COMPLETE else "failed"
+    return f"Transfer {status_text}. You can check the transaction on [blockchain explorer]({tx_link})"
+
+
 def decode_custom_error(contract_abi, w3, error) -> Tuple[str, str]:
     # Parse error content, the error content must look like:
     # "Call method: submitServiceProof,xxxxx,error:('xxxxxx','xxxxxxx')"
